/*
   Copyright 1999 by CodeFab, Inc.   All rights reserved.

   This notice may not be removed from this source code.
   The use and distribution of this software is governed by the
   terms of the MiscKit license agreement.  Refer to the license
   document included with the MiscKit distribution for the terms.
   Details of the license agreement may be found at
   www.misckit.com or within the MiscKit distribution.

   There may be more information about this software on
   www.codefab.com.  Send queries to ask_codefab@codefab.com.
*/

/*"
 * These macros will take an unquoted value and either place quotes around
 * it or return an instance of NSString that contains the quoted value.
 *
 * CFRStringify() can be used to turn an argument name into a string that can
 * be manipulated like any other static C string.
"*/
#define CFRStringify(_arg_)		#_arg_
#define CFRNSStringify(_arg_)	\
                 [NSString stringWithCString: CFRStringify(_arg_)]

/*"
 * Safely set object value of variable _v_ to object _o_.  Ensures that
 * retain counts are properly preserved without running the risk of either
 * causing _o_ to be dealloc'd (because it was the same object as _v_) or
 * without incurring the ineffeciencies inherit to #{-autorelease}.
"*/
// the id declaration/assignment is to prevent multiple evaluation of
// the 'argument' passed to the macro (since the argument on the right may be an expression).

#define CFRSetVariableToObject(_v_,_o_) do {					\
    id _object_ = (_o_);	       						\
    if (_v_ == _object_)							\
	break;									\
    if (_v_ != nil)							\
                [_v_ release];				     			\
    _v_ = _object_;							    	\
    if (_v_ != nil)					    		\
	[_v_ retain];					       		\
} while(0)

#ifdef DEBUG
  /*"
   * If DEBUG is not defined, calls exit().  If DEBUG is defined, print an
   * error message containing the _code_ and call abort().  If debugging,
   * abort() will pass control to gdb.  Otherwise, abort() dumps core.
  "*/
	#define cfrexit(_code_) \
do { \
		printf("aborting with code: %d\n", _code_); abort(); \
} while(0)
#else
  /*"
   * If DEBUG is not defined, calls exit().  If DEBUG is defined, print an
   * error message containing the _code_ and call abort().  If debugging,
   * abort() will pass control to gdb.  Otherwise, abort() dumps core.
  "*/
	#define cfrexit(_code_) exit(_code_)
#endif

/*"
 * Raises NSInvalidArgumentException if cond is not satisfied.  The
 * exception will contain the cond expression and the context (class and
 * method) in which the conditional evaluated to false.
 *
 * This define can only be used within a method's implementation (or within
 * any code where self and _cmd are defined).
"*/
#define CFRValidateMethodArgument(cond) \
	do { \
		if (! (cond)) { \
			[NSException raise:NSInvalidArgumentException \
                                           format:@"%@: Argument value unacceptable: (%s) == NO/nil/NULL/0.", \
                            [NSObject methodLabelForInstance:self andSelector:_cmd], #cond]; \
		} \
	} while (0)

/*"
 * Raises CFRInvalidVariableException if cond evaluates to false.  The
 * raised exception includes a description of the exception that contains
 * the cond and the context (class and method) in which the conditional
 * evaluated to false.
 *
 * This define can only be used within a method's implementation (or within
 * any code where self and _cmd are defined).
"*/
#define CFRValidateMethodVariable(cond) \
	do { \
		if (! (cond)) { \
			[NSException raise:CFRInvalidVariableException \
               format:@"%@: Variable value unacceptable: (%s) == NO/nil/NULL/0.", \
                            [NSObject methodLabelForInstance:self andSelector:_cmd], #cond]; \
		} \
	} while (0)

/*"
 * Raises an NSRangeException if argName is less than min or greater than
 * max.  Since it uses +methodLabelForInstance:andSelector: to generate the exception
 * name, this argument can only be used within a method (or by defining
 * self and _cmd within the scope that the macro is used).
 *
 * The raised notification will automatically document the name of the
 * argument, the method and class, and the range of valid values.
 *
 * The two argument version does not check the lower bounds of the
 * argument;  it is used to check unsigned quantities.
"*/
#define CFRValidateVariableRange(argName, min, max)		\
       do { 											\
		  if( (argName < min) || (argName > max) )	    \
				[NSException raise:NSRangeException		\
					format:@"%@: {%s}'s value (%ld) "   \
						@"not within bounds (%ld, %ld)",\
                                    [NSObject methodLabelForInstance:self andSelector:_cmd],      \
                        #argName, (long) argName, (long) min, (long)max]; 	\
	   } while(0)

#define CFRValidateVariableMax(argName, max) 			\
      do {              								\
		if(argName > max) 								\
			[NSException raise:NSRangeException 		\
				format:@"%@: {%s}'s value (%ld) "       \
						 @"greater than max (%ld)", 	\
                            [NSObject methodLabelForInstance:self andSelector:_cmd], #argName,    \
                         (long) argName, (long)max]; 				\
	  } while(0)

#define CFRMETHODLABEL [NSObject methodLabelForInstance: self andSelector: _cmd]