/*
 File: ScriptedPerformProxy

   Original Author: bbum

   Copyright 1999 by CodeFab, Inc.   All rights reserved.

   This notice may not be removed from this source code.
   The use and distribution of this software is governed by the
   terms of the MiscKit license agreement.  Refer to the license
   document included with the MiscKit distribution for the terms.
   Details of the license agreement may be found at
   www.misckit.com or within the MiscKit distribution.

   There may be more information about this software on
   www.codefab.com.  Send queries to ask_codefab@codefab.com.
*/

#import <Foundation/Foundation.h>
#import <EOControl/EOControl.h>
#import <EOAccess/EOAccess.h>
#import <WebObjects/WebObjects.h>

#import "ScriptedPerformProxy.h"

#import "Mssh.h"

@implementation ScriptedPerformProxy
+ scriptedPerformProxyWithObject: anObject;
{
    return [[[self alloc] initWithObject: anObject] autorelease];
}

- initWithObject: anObject;
{
    [super init];

    target = [anObject retain];

    invokeViaScriptedEvaluator = [anObject respondsToSelector: @selector(evaluateMethodNamed:count:arguments:)];

    return self;
}

- (void) dealloc;
{
    [target release];
    [super dealloc];
}

- (BOOL) respondsToSelector: (SEL) aSelector
{
    return [super respondsToSelector: aSelector];
}

- (NSMethodSignature *)methodSignatureForSelector:(SEL)aSelector;
{
    NSMethodSignature *mS = [super methodSignatureForSelector: aSelector];

    if (mS == nil)
        mS = [target methodSignatureForSelector: aSelector];

    return mS;
}

- (void)forwardInvocation:(NSInvocation *)anInvocation;
{
    NSString *methodName;
    NSMethodSignature *methodSignature;
    unsigned argumentIndex, argumentMaxIndex;
    id *arguments;  // always an array of objects to a scripted object!
    id returnValue;

    if (invokeViaScriptedEvaluator == NO) {
        [self sPP_setLastException:  nil];
        NS_DURING
            [anInvocation invokeWithTarget: target];
        NS_HANDLER
            [self sPP_setLastException: localException];
            [localException raise];
        NS_ENDHANDLER
        return;
    }

    methodSignature = [anInvocation methodSignature];
    argumentIndex = 2; // start at first ARGUMENT
    argumentMaxIndex = [methodSignature numberOfArguments];

    if (argumentIndex == argumentMaxIndex) { // only has self, _cmd....
        // don't know if this is necessary
        arguments = alloca(1 * sizeof(id));
        arguments[0] = nil;
    } else {
        // convert args to NS*s of some kind or another
        arguments = alloca( (argumentMaxIndex - argumentIndex) * sizeof(id));
        while (argumentIndex < argumentMaxIndex) {
            const char *argumentTypeString;
            unsigned argumentArrayIndex = argumentIndex - 2;
            id objectValue;
            NSObjCValue argumentValue;
            void *argumentValueLocation = &argumentValue.value;
            
            argumentTypeString = [methodSignature getArgumentTypeAtIndex: argumentIndex];
            switch (argumentTypeString[0]) {
               case NSObjCCharType:
                   [anInvocation getArgument:argumentValueLocation atIndex: argumentIndex];
                   objectValue = [NSNumber numberWithChar: argumentValue.value.charValue];
                   break;
               case NSObjCShortType:
                   [anInvocation getArgument:argumentValueLocation atIndex: argumentIndex];
                   objectValue = [NSNumber numberWithShort: argumentValue.value.shortValue];
                   break;
               case NSObjCLongType:
                   [anInvocation getArgument:argumentValueLocation atIndex: argumentIndex];
                   objectValue = [NSNumber numberWithLong: argumentValue.value.longValue];
                   break;
               case NSObjCLonglongType:
                   [anInvocation getArgument:argumentValueLocation atIndex: argumentIndex];
                   objectValue = [NSNumber numberWithLongLong: argumentValue.value.longlongValue];
                   break;
               case NSObjCFloatType:
                   [anInvocation getArgument:argumentValueLocation atIndex: argumentIndex];
                   objectValue = [NSNumber numberWithFloat: argumentValue.value.floatValue];
                   break;
               case NSObjCDoubleType:
                   [anInvocation getArgument:argumentValueLocation atIndex: argumentIndex];
                   objectValue = [NSNumber numberWithDouble: argumentValue.value.doubleValue];
                   break;
               case NSObjCSelectorType:
                   [anInvocation getArgument:argumentValueLocation atIndex: argumentIndex];
                   objectValue = NSStringFromSelector(argumentValue.value.selectorValue);
                   break;
               case NSObjCObjectType:
                   [anInvocation getArgument:argumentValueLocation atIndex: argumentIndex];
                   objectValue = argumentValue.value.objectValue;
                   break;
                case NSObjCStringType:
                    [anInvocation getArgument:argumentValueLocation atIndex: argumentIndex];
                    objectValue = [NSString stringWithCString: argumentValue.value.cStringLocation];
                  break;
                case NSObjCStructType:
                case NSObjCPointerType:
                case NSObjCArrayType:
                case NSObjCUnionType:
                case NSObjCBitfield:
                case NSObjCNoType:
                case NSObjCVoidType:
                    [NSException raise: NSInvalidArgumentException
                                format: @"ScriptedPerformProxy does not support arguments with objc type of '%s'.", argumentTypeString];
                    break;                    
            }

            arguments[argumentArrayIndex] = objectValue;
            
            argumentIndex++;
        }
    }

    methodName = NSStringFromSelector([anInvocation selector]);

    NS_DURING
        [self sPP_setLastException:  nil];
        returnValue = [target evaluateMethodNamed: methodName count: (argumentMaxIndex - 2) arguments: arguments];
    NS_HANDLER
        [self sPP_setLastException: localException];
        [localException raise];
    NS_ENDHANDLER

    if (returnValue != nil) {
        const char *returnTypeString = [methodSignature methodReturnType];
        NSObjCValue returnObjCValue;
        void *returnValueLocation = &returnObjCValue.value;

        returnObjCValue.value.longlongValue = 0LL; // bogus -- longlong appears to be longest member of union....

        switch (returnTypeString[0]) {
            case NSObjCNoType:
            case NSObjCVoidType:
                // do nothing
                break;
            case NSObjCCharType:
                returnObjCValue.value.charValue = [returnValue charValue];
                break;
            case NSObjCShortType:
                returnObjCValue.value.shortValue = [returnValue shortValue];
                break;
            case NSObjCLongType:
                returnObjCValue.value.longValue = [returnValue longValue];
                break;
            case NSObjCLonglongType:
                returnObjCValue.value.longlongValue = [returnValue longLongValue];
                break;
            case NSObjCFloatType:
                returnObjCValue.value.floatValue = [returnValue floatValue];
                break;
            case NSObjCDoubleType:
                returnObjCValue.value.doubleValue = [returnValue doubleValue];
                break;
            case NSObjCSelectorType:
                returnObjCValue.value.selectorValue = NSSelectorFromString(returnValue);
                break;
            case NSObjCObjectType:
                returnObjCValue.value.objectValue = returnValue;
                break;
            case NSObjCStringType:
#warning THIS implies that the (char *) is not important-- the contents are.  HOWEVER, the contents will effectively be "autoreleased" as a result of using -cString.  HOWEVER2, any other implementation will likely lead to leaks.
                returnObjCValue.value.cStringLocation = (char *) [returnValue cString];
                break;
            case NSObjCStructType:
            case NSObjCPointerType:
            case NSObjCArrayType:
            case NSObjCUnionType:
            case NSObjCBitfield:
                [NSException raise: NSInvalidArgumentException
                            format: @"ScriptedPerformProxy does not support return values of objc type of '%s'.", returnTypeString];
                break;
        }

        [anInvocation setReturnValue: returnValueLocation];
    }
}

- (NSException *) sPP_lastException;
{
    return sPP_lastException;
}

- (void) sPP_setLastException: (NSException *) anException;
{
    if (anException != sPP_lastException) {
        [sPP_lastException release];
        sPP_lastException = [anException retain];
    }
}
@end
